/*
 * Platform dependent support for IO probing.
 *
 * Copyright (c) 2000-2003 Silicon Graphics, Inc.  All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of version 2 of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it would be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * 
 * Further, this software is distributed without any warranty that it is 
 * free of the rightful claim of any third person regarding infringement 
 * or the like.  Any license provided herein, whether implied or 
 * otherwise, applies only to this software file.  Patent licenses, if 
 * any, provided herein do not apply to combinations of this program with 
 * other software, or any other product whatsoever.
 * 
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, write the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 * 
 * Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, 
 * Mountain View, CA  94043, or:
 * 
 * http://www.sgi.com 
 * 
 * For further information regarding this notice, see: 
 * 
 * http://oss.sgi.com/projects/GenInfo/NoticeExplan
 */

#include <asm/sn/sn_sal.h>

/**
 * ia64_sn_probe_io_slot - test a memory location for readability
 * @paddr: physical address to probe
 * @size: number bytes to read (1,2,4,8)
 * @data_ptr: address to store value read by probe (-1 returned if probe fails)
 *
 * This function will probe a physical address to determine if
 * the address can be read. If reading the address causes a BUS
 * error, an error is returned. If the probe succeeds, the contents 
 * of the memory location is returned.
 *
 * Return values:
 *  0 - probe successful
 *  1 - probe failed (generated MCA)
 *  2 - Bad arg
 * <0 - PAL error
 */
u64
ia64_sn_probe_io_slot(long paddr, long size, void *data_ptr)
{
	struct ia64_sal_retval isrv;

	SAL_CALL(isrv, SN_SAL_PROBE, paddr, size, 0, 0, 0, 0, 0);

	if (data_ptr) {
		switch (size) {
			case 1:
				*((u8*)data_ptr) = (u8)isrv.v0;
				break;
			case 2:
				*((u16*)data_ptr) = (u16)isrv.v0;
				break;
			case 4:
				*((u32*)data_ptr) = (u32)isrv.v0;
				break;
			case 8:
				*((u64*)data_ptr) = (u64)isrv.v0;
				break;
			default:
				isrv.status = 2;	
		}
	}

	return isrv.status;
}
