/* 
 * Copyright (C) 2000 David J. Mckay (david.mckay@st.com)
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See linux/COPYING for more information.                            
 *
 */

#ifndef __FPGA_OD_H__
#define __FPGA_OD_H__

/* This routine will program up the fpga which interfaces to the galileo */
int init_overdrive_fpga(void);

#endif
