/*
 *  linux/include/asm-ppc/hdreg.h
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

/*
 *  This file contains the ppc architecture specific IDE code.
 */

#ifndef __ASMPPC64_HDREG_H
#define __ASMPPC64_HDREG_H

typedef unsigned long ide_ioreg_t;

#endif /* __ASMPPC64_HDREG_H */

